from z3 import *
import re
import sys

content = open('./EzLUTs_top_synth.v', 'r').read()
regex = r"LUT(\d)(.*?);"
matches = re.finditer(regex, content, re.MULTILINE | re.DOTALL)

# Parse all LUT modules in the verilog file
def parse_lut(lut_cont, num):
    rx_init = r"\.INIT\((\d+)\'h(\d+)\)"
    m_init = re.search(rx_init, lut_cont)

    obj = {'num': num}
    # How many bits is the mask for the init value
    obj['bits'] = int(m_init.group(1))
    # What is the init value
    obj['mask'] = int(m_init.group(2), 16)

    rx_inps = r"\.I(\d)\((.*?)\),"
    m_inps = re.finditer(rx_inps, lut_cont, re.MULTILINE)
    # Input number `x` is the variable named `y`
    res_inps = {}
    for x in m_inps:
        res_inps[int(x.group(1))] = x.group(2)

    # These are the inputs that we have
    obj['inps'] = res_inps

    # This is the name of the output variable
    rx_out = r"\.O\((.*?)\)"
    m_out = re.search(rx_out, lut_cont)
    obj['out'] = m_out.group(1)
    return obj

luts = []
for m in matches:
    luts.append(parse_lut(m.group(2), m.group(1)))
    # break

# Reach an lut by the name of its output variable
lut_by_name = {}
for lut in luts:
    lut_by_name[lut['out']] = lut

# Get a Z3 boolean variable based on its name
# If we already have the variable then return it, otherwise construct a new one
sym_vars = {}
def get_var(name):
    if name not in sym_vars: sym_vars[name] = Bool(name)
    return sym_vars[name]

s = Solver()

# Give a logic function that tests if a specific configuration of inputs holds
def give_conjunction(bitmask, inps, bits):
    ands = []
    ptr = 1
    cnt = 0
    while ptr < bits:
        if bitmask & ptr: ands.append(get_var(inps[cnt]))
        else: ands.append(Not(get_var(inps[cnt])))

        ptr <<= 1
        cnt += 1

    return And(ands)

# Init values for LUT6 that only check on specific configuration of the 6 input values
lut6_defaults = [
4294967296,
68719476736,
2251799813685248,
9223372036854775808,
1152921504606846976,
2048,
2,
576460752303423488,
34359738368,
549755813888,
16,
32,
]

# Init values for LUT6 that are not single configuration checks
# The comments on the right display the indices (0 based) of the bits that are set (i.e equal to one) in the init value on the left side
# Based on the indices where the output is 1, and looking at the truth tables in the AMD documentation page we manually figure out which simple logic function describes the LUT fully
# The parentheses after the indices shortly describe the logic function that I arrived to, and the lambdas return a Z3 value based on the inputs that realize the given function
lut6_custom = {
    116094737448960: lambda i0,i1,i2,i3,i4,i5: And(i0 ^ i1 ^ i2 ^ i3, i4 ^ i5), # - 17,18,20,23,24,27,29,30,33,34,36,39,40,43,45,46, (not xor bits, and xor i4, i5)
    10838310071209820160: lambda i0,i1,i2,i3,i4,i5: And(i0 ^ i1 ^ i2 ^ i3 ^ i4, i5), #- 33,34,36,39,40,43,45,46,48,51,53,54,57,58,60,63, (not xor bits and i5)
    10808639749944153600: lambda i0,i1,i2,i3,i4,i5: And(i0 ^ i1 ^ i2, i3 ^ i4 ^ i5), # - 9,10,12,15,17,18,20,23,33,34,36,39,57,58,60,63, (not xor bits and Xor(i3,i4,i5))
    6917529439957942272: lambda i0,i1,i2,i3,i4,i5: And(i5, i2, Not(Xor(i4,i3)), Xor(i0,i1)), # - 37,38,61,62, (i5 and i2 and notxor(i4,i3) and xor(i0,i1))
    16660: lambda i0,i1,i2,i3,i4,i5: And(i1 ^ i2 ^ i3, Not(i0), Not(i4), Not(i5)), # - 2,4,8,14, (i1,i2,i3 xor others are null)
    2523490710: lambda i0,i1,i2,i3,i4,i5: And(i0 ^ i1 ^ i2 ^ i3 ^ i4, Not(i5)), # - 1,2,4,7,8,11,13,14,16,19,21,22,25,26,28,31, (xor everything and Not(i5))
    42386175767347350: lambda i0,i1,i2,i3,i4,i5: And(i0 ^ i1 ^ i2, Not(i3 ^ i4 ^ i5)), # - 1,2,4,7,25,26,28,31,41,42,44,47,49,50,52,55, (xor everything and Not(Xor(i4,i3,i5)))
    7608268620489058710: lambda i0,i1,i2,i3,i4,i5: And(i0 ^ i1 ^ i2 ^ i3, Not(i4 ^ i5)), # - 1,2,4,7,8,11,13,14,49,50,52,55,56,59,61,62, (xor everything and Not(Xor(i4,i5)))
    7608434000728254870: lambda i0,i1,i2,i3,i4,i5: i0 ^ i1 ^ i2 ^ i3 ^ i4 ^ i5, # - 1,2,4,7,8,11,13,14,16,19,21,22,25,26,28,31,32,35,37,38,41,42,44,47,49,50,52,55,56,59,61,62, (xor everything)
    29670323037143145: lambda i0,i1,i2,i3,i4,i5: And(Not(i0 ^ i1 ^ i2), Not(i3 ^ i4 ^ i5)), # - 0,3,5,6,24,27,29,30,40,43,45,46,48,51,53,54, (not xor bits and not xor i4,i3,i5)
    1771476585: lambda i0,i1,i2,i3,i4,i5: And(Not(i0 ^ i1 ^ i2 ^ i3 ^ i4), Not(i5)), # - 0,3,5,6,9,10,12,15,17,18,20,23,24,27,29,30, (not xor bits and not i5)
    10838193978243847785: lambda i0,i1,i2,i3,i4,i5: And(Not(i0 ^ i1 ^ i2 ^ i3), Not(i4 ^ i5)) # - 0,3,5,6,9,10,12,15,48,51,53,54,57,58,60,63, (not xor bits and not xor i4,i5)
}

# Visit the tree of dependency in depth first order
def dfs(cur):
    # Base case, just return a symbolic value for the unknown flag bit
    if 'data_IBUF' in cur: return get_var(cur)

    cur_lut = lut_by_name[cur]
    # Compute the values of all inputs (symbolic based on input ibuf variables)
    for inp in cur_lut['inps'].values():
        dfs(inp)

    # Set the value of the current variable, as determined by the simplified LUT function
    if cur_lut['num'] == '2': sym_vars[cur] = get_var(cur_lut['inps'][0]) ^ get_var(cur_lut['inps'][1])
    elif cur_lut['num'] == '3': sym_vars[cur] = get_var(cur_lut['inps'][0]) ^ get_var(cur_lut['inps'][1]) ^ get_var(cur_lut['inps'][2])
    elif cur_lut['num'] == '4': sym_vars[cur] = get_var(cur_lut['inps'][0]) ^ get_var(cur_lut['inps'][1]) ^ get_var(cur_lut['inps'][2]) ^ get_var(cur_lut['inps'][3])
    elif cur_lut['num'] == '5':
        if cur_lut['mask'] == 536870912 or cur_lut['mask'] == 134217728: sym_vars[cur] = give_conjunction(bin(cur_lut['mask'])[2:][::-1].index('1'), cur_lut['inps'], cur_lut['bits'])
        else: sym_vars[cur] = get_var(cur_lut['inps'][0]) ^ get_var(cur_lut['inps'][1]) ^ get_var(cur_lut['inps'][2]) ^ get_var(cur_lut['inps'][3]) ^ get_var(cur_lut['inps'][4])
    else:
        if cur_lut['mask'] in lut6_defaults: sym_vars[cur] = give_conjunction(bin(cur_lut['mask'])[2:][::-1].index('1'), cur_lut['inps'], cur_lut['bits'])
        else: sym_vars[cur] = lut6_custom[cur_lut['mask']](*[get_var(cur_lut['inps'][x]) for x in range(6)])

# We know the flag byte starting at idx (idx specified in bits)
def set_byte(idx, val):
    for i in range(8):
        exp = True if (val>>(8-i-1))&1 == 1 else False
        var = get_var(f'data_IBUF[{idx+i}]')
        print('constraint: ', f'data_IBUF[{idx+i}]', '==', exp)
        s.add(var == exp)

# Compute the logic function for the success output
dfs('success_OBUF')

# idx 0 within current byte is the highest bit of the byte
# bytes are left-to-right as expected
# i.e highest bit of first byte is index 0
# lower bit of first byte is index 7
for x in range(0, 336, 8):
    var0 = get_var(f'data_IBUF[{x}]')
    var1 = get_var(f'data_IBUF[{x+1}]')
    var2 = get_var(f'data_IBUF[{x+2}]')
    # Enforce byte is >= 32
    s.add(Or(var1, var2) == True)
    # Enforce byte < 128
    s.add(var0 == False)

# Flag prefix and suffix
set_byte(0, ord('0'))
set_byte(1*8, ord('o'))
set_byte(2*8, ord('p'))
set_byte(3*8, ord('s'))
set_byte(4*8, ord('{'))
set_byte(336-8, ord('}'))

# Success of flag check should be true
s.add(get_var('success_OBUF') == True)

# Check if model is satisfiable
print(s.check())

# Convert flag bit symbolic values to actual flag
mod = s.model()
flag = ''
for i in range(0, 336, 8):
    v = 0
    idk = False
    for j in range(8):
        name = f'data_IBUF[{i+j}]'
        val = str(mod.evaluate(get_var(name)))
        v |= (1<<(8-j-1)) if val == 'True' else 0

    flag += chr(v)

    print('byte', i//8, v, chr(v), idk)

print('flag:', flag)
